// License: GPLv3 or later
// Author: AdmSasha <dik@inbox.ru>

#include "cBuild.h"
#include "functions.h"
#include "cChroot.h"

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>

#include <fstream>

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/asio/signal_set.hpp>
#include <boost/asio/impl/io_context.hpp>
#include <boost/thread.hpp>

using namespace boost::filesystem;

cBuild::cBuild(){
}

int cBuild::build(map<string,string> _options){
    sProcResult procresult;
    isDeleteTmp=false;
    procBuildPid=0;
    normalExit=false;
    stopped=false;
    int status=0;

    options=_options;
    
    // Forced change group (Принудительная смена группы)
    pid_t mypid=getpid();
    setpgid(mypid,mypid);
    
    // Check options (Проверка options)
    if (options["upload-dir"]=="") options["upload-dir"]=".";
    if (options["tmp-dir"]=="") options["tmp-dir"]="~/tmp/aum++-"+rndString(8);
    if (options["timeout"]=="") options["timeout"]="0";

    options["chroot_dir"]=options["tmp-dir"]+"/chroot";
    
    string urpmi_options=replaceString(options["urpmi-options"],","," ");

    // Check the build settings (Проверка параметров сборки)
    if (checkOptions()!=0) return 1;

    // Use the interrupt handler (Ставим обработчик прерываний)
    boost::asio::signal_set sig(service, SIGINT, SIGTERM);
    sig.async_wait(boost::bind(&cBuild::signalExit, this));
    boost::thread thr(boost::bind(&boost::asio::io_service::run, &service));

    // Check directory tmp-dir (Проверяем папку tmp-dir)
    if (boost::filesystem::exists(options["tmp-dir"])==false){
	isDeleteTmp=true;
    }

    if (boost::filesystem::exists("~/tmp")==false) pbsMkdir("~/tmp");

    logdebug(_("Temporary directory (tmp-dir): ")+options["tmp-dir"],options);

    
    
    
    cChroot cc(options);

    // Create chroot (Создание chroot)
    // Unpack the archive (Распаковываем архив)
    if (options["chroot"]!=""){
        logdebug(_("Unpacking chroot archive..."),options);
        procresult=unPackArchive();
        //logdebug("tar code result: "+IntToString(procresult),options);
	if (procresult.status!=0){
	    logdebug(CharToString(_("Error when unpacking %s in %s"),options["chroot"].c_str(),options["tmp-dir"].c_str()),options);
    	    clearBuild();
	    return 1;
	}
	logdebug(_("Unpacking chroot archive was completed"),options);
	if (stopped==true){clearBuild(); return 1;}
    }
    
    if (boost::filesystem::exists(options["chroot_dir"])==false){
	logdebug(CharToString(_("%s not found"),options["chroot_dir"].c_str()),options);
	clearBuild();
	return 1;
    }

    logdebug(_("Mount dev "),options);
    cc.mountDev();
    if (stopped==true){clearBuild(); return 1;}

    
    string chroot_dir=options["chroot_dir"];
    
    // Create a user and his environment (Создание пользователя и его окружения)
    logdebug(CharToString(_("Create user with name: %s"),buildUser().c_str()),options);
    createUser();
    if (stopped==true){clearBuild(); return 1;}

    logdebug(_("Copy addons "),options);
    copyAddons();
    if (stopped==true){clearBuild(); return 1;}
    
    //logdebug(_("Add distrib "),options);
    status=addDistrib();
    if (stopped==true or status!=0){clearBuild(); return 1;}

    status=addDistribOwn();
    if (stopped==true or status!=0){clearBuild(); return 1;}

    //logdebug(_("Add media "),options);
    status=addMedia();
    if (stopped==true or status!=0){clearBuild(); return 1;}
    
    
    logdebug(CharToString(_("Copying srpm into chroot_dir/home/%s"),buildUser().c_str()),options);
    copySrpm();
    

    //std::cout << "Rebuilding SRPM..." << std::endl;
    //exec_cmd_in_chroot("'rpm -ivh /home/aum/$srpm_basename'",            $chroot_dir, user => 'aum');
    //exec_cmd_in_chroot("'rpmbuild -bs /home/aum/rpmbuild/SPECS/*.spec'", $chroot_dir, user => 'aum');

    runproc("chown "+buildUser()+":"+buildUser()+" /home/"+buildUser()+" -R",true,"",0);
    
    
    string srpm_filepath="/home/"+buildUser()+"/srpms/"+basename(options["srpm"]);
    
    logdebug(_("Rebuilding RPM database..."),options);
    procresult=runproc("rpm --rebuilddb",true,"",0);
    if (stopped==true or procresult.status!=0){clearBuild(); return 1;}

    logdebug(_("Installing build dependencies..."),options);
    procresult=runproc("urpmi --buildrequires --debug --no-verify-rpm --nolock --auto --ignoresize --no-recommends "+urpmi_options+" "+srpm_filepath, true,"",0);
    if (stopped==true or procresult.status!=0){clearBuild(); return 1;}


    // Building (Сборка)
    logdebug(_("Building..."),options);
    sProcResult resultBuild;
    if (options["arch"] == "i586") {
	resultBuild=runproc("'linux32 rpmbuild --target i586 -ba /home/"+buildUser()+"/rpmbuild/SPECS/*'",true,buildUser(),atoi(options["timeout"].c_str()));
	//resultBuild=runproc("'linux32 rpmbuild --target i586 --rebuild "+srpm_filepath+"'",true,buildUser(),atoi(options["timeout"].c_str()));
    }else {
	resultBuild=runproc("'rpmbuild -ba /home/"+buildUser()+"/rpmbuild/SPECS/*'",true,buildUser(),atoi(options["timeout"].c_str()));
        //resultBuild=runproc("'rpmbuild --rebuild "+srpm_filepath+"'",true,buildUser(),atoi(options["timeout"].c_str()));
    }
    if (stopped==true or resultBuild.status!=0){clearBuild(); return 1;}
    
    
    // Copy files after building into destination (Раскидываем собранные файлы)
    logdebug(_("Copy RPM and SRPM packages into upload-dir..."),options);
    status = copyBuiltPackages();
    if (status!=0){clearBuild(); return 1;}
    
    
    
    // Run rpmlint (Проверка rpmlint)
    if (options["norpmlint"]!="true"){
        if (boost::filesystem::exists(chroot_dir+"/usr/bin/rpmlint")==true){
	    vector<string> listfiles = getRecurciveListFiles(chroot_dir+"/home/"+buildUser()+"/rpmbuild/RPMS");
	    for (unsigned int i=0;i<listfiles.size();i++){
		logdebug(CharToString(_("rpmlint %s"),basename(listfiles[i]).c_str()),options);
		runproc("rpmlint "+listfiles[i].substr(chroot_dir.length()), true,"",0);
    	    }
    	    
	    listfiles = getRecurciveListFiles(chroot_dir+"/home/"+buildUser()+"/rpmbuild/SRPMS");
	    for (unsigned int i=0;i<listfiles.size();i++){
		logdebug(CharToString(_("rpmlint %s"),basename(listfiles[i]).c_str()),options);
		runproc("rpmlint "+listfiles[i].substr(chroot_dir.length()), true,"",0);
    	    }
    	    
        }else{
    	    logdebug("/usr/bin/rpmlint not found",options);
        }
    }
    
    
        
    clearBuild();
    
    
    return 0;
}

void cBuild::signalExit(){
    if (normalExit==true) return;
    std::cout << _("\nManual exit\n");
    stopped=true;
}

void cBuild::clearBuild(){
    // Unmount /dev (Размонтируем /dev)
    //cc.umountDev();
    service.stop();
    
    if (procBuildPid>0){
	kill(procBuildPid,15);
	procBuildPid=0;
	wait(3);
    }
    
    string chroot_dir=options["chroot_dir"];
    
    
    for (size_t i=0;i<list_mount_dev.size();i++){
	if (isMount(list_mount_dev.at(i))==true){
	    string cmd="umount "+list_mount_dev.at(i);
	    system(cmd.c_str());
	}
    }
    

    if (isMount(chroot_dir+"/dev/pts")==true){
	string cmd="umount "+chroot_dir+"/dev/pts";
	system(cmd.c_str());
    }
    if (isMount(chroot_dir+"/dev")==true){
	string cmd="umount "+chroot_dir+"/dev";
	system(cmd.c_str());
    }
    if (isMount(chroot_dir+"/sys")==true){
	string cmd="umount "+chroot_dir+"/sys";
	system(cmd.c_str());
    }
    if (isMount(chroot_dir+"/proc")==true){
	string cmd="umount "+chroot_dir+"/proc";
	system(cmd.c_str());
    }
    
    // Delete temporary output file (Удаление временного файла вывода)
    FileRemove(options["tmp-dir"]+"/aum++_proc_output_last.tmp");
    
    // Delete temporary directory (Удаляем временную папку)
    if (options["noclear"]!="true"){
	try {
	    logdebug(_("Delete directory ")+options["tmp-dir"]+"/chroot",options);
	    FolderRemove(options["tmp-dir"]+"/chroot");
	    if (isDeleteTmp==true){
		logdebug(_("Delete directory ")+options["tmp-dir"],options);
		boost::filesystem::remove_all(options["tmp-dir"]);
	    }
        }catch (boost::filesystem::filesystem_error &e){ 
	    std::cerr << e.what() << std::endl; 
	} 
    }
    
    normalExit=true;
}


sProcResult cBuild::unPackArchive(){
    string tmp_dir=options["tmp-dir"];
    string chroot=options["chroot"];
    boost::smatch xResults;
    sProcResult result;
    
    if (boost::filesystem::exists(tmp_dir)==false){
	pbsMkdir(tmp_dir.c_str());
    }
    
    if (boost::regex_match(chroot, xResults, boost::regex("(.*)(\\.tar\\.gz)$"))==true){
	result=runproc("tar zxf "+chroot+"  --absolute-names -C "+tmp_dir,false,"",0);
    }
    if (boost::regex_match(chroot, xResults, boost::regex("(.*)(\\.tar\\.bz2)$"))==true){
	result=runproc("tar jxf "+chroot+"  --absolute-names -C "+tmp_dir,false,"",0);
    }
    if (boost::regex_match(chroot, xResults, boost::regex("(.*)(\\.tar\\.xz)$"))==true){
	result=runproc("tar Jxf "+chroot+"  --absolute-names -C "+tmp_dir,false,"",0);
    }

    return result;
}


int cBuild::createUser(){
    string chroot_dir=options["chroot_dir"];
    //mkdir ~/.gnupg; touch ~/.gnupg/gpg-agent-no-start

    runproc("userdel -fr "+buildUser(),true,"",0);
    FolderRemove(chroot_dir+"/home/"+buildUser());
    runproc("useradd "+buildUser(),true,"",0);

    pbsMkdir(chroot_dir+"/home/"+buildUser()+"/.gnupg");
    touchFile(chroot_dir+"/home/"+buildUser()+"/.gnupg/gpg-agent-no-start");

    runproc("chown -R "+buildUser()+":"+buildUser()+" /home/"+buildUser(),true,"",0);

    return 0;
}

int cBuild::copyAddons(){
    string chroot_dir=options["chroot_dir"];

    if (copyFile("/etc/resolv.conf",chroot_dir+"/etc/resolv.conf")!=0){
        std::cout << _("Can't copy /etc/resolv.conf into chroot") << std::endl;
    }

    if (options["add-rpmmacros"] != ""){
	copyFile(options["add-rpmmacros"],chroot_dir+"/home/"+buildUser()+"/.rpmmacros");
    }

    return 0;
}

int cBuild::addDistrib(){
    string chroot_dir=options["chroot_dir"];
    
    if (options["distrib"]=="") return 0;
    
    vector<string> resultSplit;
    boost::split(resultSplit,options["distrib"],boost::is_any_of(";"),boost::token_compress_on);
    for (unsigned int i=0;i<resultSplit.size();i++){
	string distrib=resultSplit[i];
	if (distrib=="") continue;
	logdebug(_("Add distrib: ")+ distrib,options);
	
	
    	boost::smatch xDistrib;
	if (boost::regex_match(distrib,xDistrib,boost::regex("^(file://)(.*)"))==true){
	    std::string bindname=chroot_dir+"/media/distrib_"+IntToString(i);
	    pbsMkdir(bindname);
	    pbsMount(xDistrib[2],bindname);
	    list_mount_dev.push_back(bindname);
	    
	    distrib="/media/distrib_"+IntToString(i);
        }
	
	sProcResult procResult=runproc("urpmi.addmedia --distrib "+distrib,true,"",0);
        if (procResult.status!=0) return 1;
	// Check output (Проверка вывода)
	string path_output_tmp = options["tmp-dir"]+"/aum++_proc_output_last.tmp";

        std::ifstream ifs(path_output_tmp.c_str());
	if (!ifs.is_open()) {
    	    std::cerr << _("Failed to open file") << std::endl;
	    return 2;
	}
        std::string buffer;
	boost::smatch xResults;
	while (std::getline(ifs, buffer)) {
	    if (boost::regex_match(buffer, xResults, boost::regex("(.*)(exited with)(.*)"))==true){
		return 1;
	    }
	}
	ifs.close();
    }
    return 0;
}

int cBuild::addDistribOwn(){
    string chroot_dir=options["chroot_dir"];
    
    if (options["distrib-own"]=="") return 0;
    
    vector<string> resultSplit;
    boost::split(resultSplit,options["distrib-own"],boost::is_any_of(";"),boost::token_compress_on);
    for (unsigned int i=0;i<resultSplit.size();i++){
	string distrib=resultSplit[i];
	if (distrib=="") continue;
	logdebug(_("Add distrib own: ")+ distrib,options);
	
	std::string mDistrib=distrib;
	
    	boost::smatch xDistrib;
	if (boost::regex_match(distrib,xDistrib,boost::regex("^(file://)(.*)(((/)(.+)){2})$"))==true){
	    std::string bindname=chroot_dir+"/media/distrib_"+IntToString(i);
	    pbsMkdir(bindname);
	    pbsMount(xDistrib[2],bindname);
	    list_mount_dev.push_back(bindname);
	    
	    distrib="/media/distrib_"+IntToString(i);
    	    mDistrib=distrib+xDistrib[3];
        }
	
	sProcResult procResult=runproc("urpmi.addmedia --distrib "+mDistrib,true,"",0);
        if (procResult.status!=0) return 1;
	// Check output (Проверка вывода)
	string path_output_tmp = options["tmp-dir"]+"/aum++_proc_output_last.tmp";

        std::ifstream ifs(path_output_tmp.c_str());
	if (!ifs.is_open()) {
    	    std::cerr << _("Failed to open file") << std::endl;
	    return 2;
	}
        std::string buffer;
	boost::smatch xResults;
	while (std::getline(ifs, buffer)) {
	    if (boost::regex_match(buffer, xResults, boost::regex("(.*)(exited with)(.*)"))==true){
		return 1;
	    }
	}
	ifs.close();
    }
    return 0;
}


int cBuild::addMedia(){
    string chroot_dir=options["chroot_dir"];
    
    if (options["add-media"]=="") return 0;
    
    vector<string> resultSplit;
    boost::split(resultSplit,options["add-media"],boost::is_any_of(";"),boost::token_compress_on);
    for (unsigned int i=0;i<resultSplit.size();i++){
	string media=resultSplit[i];
	if (media=="") continue;
	string media_name="media_"+rndString(8)+"_"+IntToString(i);
	
	std::cout << _("Add media: ") << media << std::endl;
        sProcResult procResult=runproc("urpmi.addmedia "+media_name+" "+media, true,"",0);
        
        std::cout << "Add media: " << procResult.status << endl;
        
        if (procResult.status!=0) return 1;
        if (procResult.text.length()==0) return 1;
        
        // Check errors (Проверки на ошибки)
        vector<string> resultSplit;
	boost::split(resultSplit,procResult.text,boost::is_any_of("\n"),boost::token_compress_on);
	for (size_t y=0;y<resultSplit.size();y++){
	    boost::smatch xResults;
	    if (boost::regex_match(resultSplit[y], xResults, boost::regex("(.*)(exited with)(.*)"))==true){
		return 1;
	    }
	}
	
	// Check in urpmi.cfg (Проверка в urpmi.cfg)
	string urpmi_cfg_path=chroot_dir+"/etc/urpmi/urpmi.cfg";
	std::ifstream ifs(urpmi_cfg_path.c_str());
        std::string buffer;
        boost::smatch xResults;
        bool mediaExist=0;
	if (!ifs.is_open()){
    	    std::cerr << _("Failed to open file: ") << "/etc/urpmi/urpmi.cfg" << std::endl;
            return 1;
	}
	while (std::getline(ifs, buffer)) {
	    //std::cout << buffer << endl;
	    if (boost::regex_match(buffer,  xResults, boost::regex("("+media_name+")(.*)"))==true){
		mediaExist=1;
		break;
	    }
	}
	if (mediaExist==0) return 1;
	
    }

    return 0;
}

int cBuild::copySrpm(){
    string chroot_dir=options["chroot_dir"];
    sProcResult procResult;
    
    if (options["srpm"]==""){
	std::cout << _("srpm not found") << std::endl;
	return 1;
    }
    
    // Copy src.rpm into /chroot/home/user/srpms (Копирование src.rpm в /chroot/home/user/srpms)
    string srpm_filename=basename(options["srpm"]);
    string chroot_srpm_filename=chroot_dir+"/home/"+buildUser()+"/srpms/"+srpm_filename;
    pbsMkdir(chroot_dir+"/home/"+buildUser()+"/srpms");
    copyFile(options["srpm"],chroot_srpm_filename);

    // Unpack src.rpm files for building (Расспаковка src.rpm файлов для сборки)
    string path_srpms_files=chroot_dir+"/home/"+buildUser()+"/srpms/files/";
    pbsMkdir(path_srpms_files);
    procResult=runproc("cd "+path_srpms_files+" && rpm2cpio "+chroot_srpm_filename+" | cpio -imvu --no-absolute-filenames",false,"",0);
    
    // Sort files (Разброс файлов)
    boost::smatch xResults;
    vector<string> listfiles = getRecurciveListFiles(path_srpms_files);
    pbsMkdir(chroot_dir+"/home/"+buildUser()+"/rpmbuild/SPECS/");
    pbsMkdir(chroot_dir+"/home/"+buildUser()+"/rpmbuild/SOURCES/");
    for (unsigned int i=0;i<listfiles.size();i++){
        if (boost::regex_match(basename(listfiles[i]), xResults, boost::regex("^(.*)\\.spec$"))==true){
    	    copyFile(listfiles[i],chroot_dir+"/home/"+buildUser()+"/rpmbuild/SPECS/"+basename(listfiles[i]));
	}else{
    	    copyFile(listfiles[i],chroot_dir+"/home/"+buildUser()+"/rpmbuild/SOURCES/"+basename(listfiles[i]));
	}
    }
    runproc("chown "+buildUser()+":"+buildUser()+" /home/"+buildUser()+"/srpms/ -R",true,"",0);
    
    
    
    //boost::filesystem::remove_all(path_srpms_files);
    return 0;
}


sProcResult cBuild::runproc(string cmd,bool isChroot,string user,int timeout){
    sProcResult result;
    result.status=1;
    result.text="";
    
    string chroot_dir=options["chroot_dir"];
    int status=0;
    int isWaitKill=0;

    // Building (Сборка)
    procBuildPid = fork();
    if (procBuildPid==-1){
	std::cout << _("/* Handle the error (you can perror(\"fork\") and exit) */") << std::endl;
        stopped=true;
	return result;
    }
    if (procBuildPid==0){
	int exitcode=0;

        if (isChroot==true){
	    exitcode=exec_cmd_in_chroot(cmd,chroot_dir,user,options);
	}else{
	    exitcode=exec_cmd(cmd,options);
	}
	//std::cout << "exit code: " << exitcode << endl;
	exit(exitcode);
    }

    double starttime=time(NULL);
    int npid=0;
    while (npid != procBuildPid){
	wait(1);

        npid = waitpid(procBuildPid, &status,WNOHANG);
        if (npid == -1) {
            perror("waitpid");
            break;
        }
        if (isWaitKill==0){
	    if (time(NULL)-starttime>timeout and timeout!=0){
		logdebug(_("\nError: stopped by timeout"),options);
		isWaitKill=1;
		kill(0,15);
		wait(2);
	    }
	    if (stopped==true){
		isWaitKill=1;
		kill(0,15);
		wait(2);
	    }
	}
    }
    procBuildPid=0;
    if (npid == -1) return result;
    
    if (WIFEXITED(status)) {
	//printf("exited, status=%d\n", WEXITSTATUS(status));
	status=WEXITSTATUS(status);
    } else if (WIFSIGNALED(status)) {
	//printf("killed by signal %d\n", WTERMSIG(status));
	status=WTERMSIG(status);
    } else if (WIFSTOPPED(status)) {
        //printf("stopped by signal %d\n", WSTOPSIG(status));
        status=WSTOPSIG(status);
    } else if (WIFCONTINUED(status)) {
	//printf("continued\n");
	status=1;
    }
    
    result.status=status;
    result.text+="1";
    
    return result;
}


int cBuild::checkOptions(){

    // Check chroot tarball if there is no parameter (Проверка существования chroot образа, если указан)
    if (options["chroot"]!=""){
	if (boost::filesystem::exists(options["chroot"])==false){
	    logdebug(CharToString(_("%s not found"),options["chroot"].c_str()),options);
	    return 1;
	}
    }

    // Check srpm (Проверка srpm)
    if (boost::filesystem::exists(options["srpm"])==false){
        logdebug(CharToString(_("%s not found"),options["srpm"].c_str()),options);
	return 1;
    }

    
    // Check rpmmacros if there is parameter (Проверка существования rpmmacros, если указан)
    if (options["add-rpmmacros"]!=""){
	if (boost::filesystem::exists(options["add-rpmmacros"])==false){
	    logdebug(CharToString(_("%s not found"),options["add-rpmmacros"].c_str()),options);
	    return 1;
	}
    }

    return 0;
}

int cBuild::copyBuiltPackages(){
    string chroot_dir=options["chroot_dir"];
    
    if (boost::filesystem::exists(options["upload-dir"])==false){
	pbsMkdir(options["upload-dir"]);
    }
    if (options["upload-rpm-dir"]!=""){
        if (boost::filesystem::exists(options["upload-rpm-dir"])==false){
	    pbsMkdir(options["upload-rpm-dir"]);
	}
    }
    if (options["upload-srpm-dir"]!=""){
	if (boost::filesystem::exists(options["upload-srpm-dir"])==false){
	    pbsMkdir(options["upload-srpm-dir"]);
	}
    }
    
    vector<string> listfiles = getRecurciveListFiles(chroot_dir+"/home/"+buildUser()+"/rpmbuild/RPMS");
    if (listfiles.size()==0) return 1;
    for (unsigned int i=0;i<listfiles.size();i++){
	logdebug(CharToString(_("  - %s"),basename(listfiles[i]).c_str()),options);
        if (options["upload-rpm-dir"]!=""){

	    string arch="noarch";
    	    boost::smatch xResults;
	        
	    if (boost::regex_match(basename(listfiles[i]),  xResults, boost::regex("(.*)(\\.)(src|x86_64|i586)(\\.rpm)"))==true){
		arch=xResults[3];
	    }
	    string targetRepoFolder = options["upload-rpm-dir"]+"/"+arch;
	
	    if (boost::filesystem::exists(targetRepoFolder)==false){
		pbsMkdir(targetRepoFolder);
	    }
    	    copyFile(listfiles[i],targetRepoFolder+"/"+basename(listfiles[i]));
        }else{
    	    copyFile(listfiles[i],options["upload-dir"]+"/"+basename(listfiles[i]));
        }
    }
    
    listfiles = getRecurciveListFiles(chroot_dir+"/home/"+buildUser()+"/rpmbuild/SRPMS");
    for (unsigned int i=0;i<listfiles.size();i++){
	logdebug(CharToString(_("  - %s"),basename(listfiles[i]).c_str()),options);
        if (options["upload-srpm-dir"]!=""){
    	    copyFile(listfiles[i],options["upload-srpm-dir"]+"/"+basename(listfiles[i]));
        }else{
    	    copyFile(listfiles[i],options["upload-dir"]+"/"+basename(listfiles[i]));
        }
    }
    
    return 0;
}

